import pathlib

import pytest


@pytest.fixture(scope="session")
def test_dir():
    base = pathlib.Path(__file__).parent
    return base.joinpath("data")


@pytest.fixture(scope="session")
def wav_dir(test_dir):
    return test_dir.joinpath("wav")


@pytest.fixture(scope="session")
def am_dir(test_dir):
    return test_dir.joinpath("acoustic_models")


@pytest.fixture(scope="session")
def temp_dir(test_dir):
    p = test_dir.joinpath("temp")
    p.mkdir(exist_ok=True)
    return p


@pytest.fixture(scope="session")
def dictionaries_dir(test_dir):
    return test_dir.joinpath("dictionaries")


@pytest.fixture(scope="session")
def wav_path(wav_dir):
    return wav_dir.joinpath("acoustic_corpus.wav")


@pytest.fixture(scope="session")
def transition_model_path(am_dir):
    return am_dir.joinpath("final.mdl")


@pytest.fixture(scope="session")
def tree_path(am_dir):
    return am_dir.joinpath("tree")


@pytest.fixture(scope="session")
def dictionary_path(dictionaries_dir):
    return dictionaries_dir.joinpath("test_basic.txt")



from _kalpy.hmm import TransitionModel
from _kalpy.util import ReadKaldiObject


def test_transition_model(transition_model_path):
    tm = TransitionModel()
    ReadKaldiObject(str(transition_model_path), tm)
    assert len(tm.GetPhones()) > 0



from __future__ import annotations

import dataclasses
import pathlib
import typing

import librosa
import numpy as np

from _kalpy import feat
from _kalpy.matrix import CompressedMatrix, FloatMatrixBase
from _kalpy.util import (
    BaseFloatMatrixWriter,
    CompressedMatrixWriter,
    RandomAccessBaseFloatMatrixReader,
    SequentialBaseFloatMatrixReader,
)


@dataclasses.dataclass
class Segment:
    """
    Data class for information about acoustic segments
    """

    file_path: str
    begin: typing.Optional[float] = None
    end: typing.Optional[float] = None
    channel: typing.Optional[int] = 0


class MfccComputer:
    """
    Class for computing MFCC features

    Parameters
    ----------
    sample_frequency: float
        Sample rate to use in generating features, audio will be resampled as necessary,
        defaults to 16000
    frame_length: int
        Frame length in milliseconds, defaults to 25
    frame_shift: int
        Frame shift in milliseconds, defaults to 10
    dither: float
        Dithering to use for generating deterministic features while avoiding numerical zeros,
        defaults to -1
    preemphasis_coefficient: float
        Pre-emphasis coefficient to use prior to feature calculation, defaults to 0.97
    remove_dc_offset: bool
        Flag for removing DC offset, defaults to True
    window_type: str
        Type of window to use in generating frames, defaults to "povey"
    round_to_power_of_two: bool
        Flag for using a window based on the power of two in FFT calculation for efficiency,
        defaults to True
    blackman_coeff: float
        Coefficient to use when `window_type` is "blackman", defaults to 0.42
    snip_edges: bool
        Flag for whether edges of segments should be cutoff to ensure no zero padding in frames,
        defaults to True
    max_feature_vectors: int
        Maximum number of vectors to store in memory for VTLN calculation, defaults to -1
    num_mel_bins: int
        Number of mel frequency bins to use in calculating MFCCs, defaults to 23
    low_frequency: float
        Lowest frequency for the mel spectrum, defaults to 20
    high_frequency: float
        Highest frequency for the mel spectrum, defaults to 7800
    vtln_low: float
        VTLN lower cutoff of warping function, defaults to 100
    vtln_high: float
        VTLN upper cutoff of warping function if negative,
        added to the Nyquist frequency to get the cutoff, defaults to -500
    num_coefficients: int
        Number of MFCC coefficients, defaults to 13
    use_energy: bool
        Use energy of frame in place of the zeroth MFCC coefficient, defaults to True
    energy_floor: float
        Energy floor for MFCC, defaults to 0, set to 1.0 or 0.1 if dithering is disabled
    raw_energy: bool
        Flag for computing energy before pre-emphasis and windowing, defaults to True
    cepstral_lifter: float
        Scaling factor on cepstra for HTK compatibility, defaults to 22.0
    htk_compatibility: bool
        Flag for generating features in HTK format
    """

    def __init__(
        self,
        sample_frequency: float = 16000,
        frame_length: int = 25,
        frame_shift: int = 10,
        dither: float = 1.0,
        preemphasis_coefficient: float = 0.97,
        remove_dc_offset: bool = True,
        window_type: str = "povey",
        round_to_power_of_two: bool = True,
        blackman_coeff: float = 0.42,
        snip_edges: bool = True,
        max_feature_vectors: int = -1,
        num_mel_bins: int = 25,
        low_frequency: float = 20,
        high_frequency: float = 7800,
        vtln_low: float = 100,
        vtln_high: float = -500,
        num_coefficients: int = 13,
        use_energy: bool = True,
        energy_floor: float = 0.0,
        raw_energy: bool = True,
        cepstral_lifter: float = 22.0,
        htk_compatibility: bool = False,
    ):
        self.frame_opts = feat.FrameExtractionOptions()
        self.frame_opts.frame_length_ms = frame_length
        self.frame_opts.frame_shift_ms = frame_shift
        self.frame_opts.dither = dither
        self.frame_opts.preemph_coeff = preemphasis_coefficient
        self.frame_opts.samp_freq = sample_frequency
        self.frame_opts.remove_dc_offset = remove_dc_offset
        self.frame_opts.window_type = window_type
        self.frame_opts.round_to_power_of_two = round_to_power_of_two
        self.frame_opts.blackman_coeff = blackman_coeff
        self.frame_opts.snip_edges = snip_edges
        self.frame_opts.max_feature_vectors = max_feature_vectors

        self.mel_opts = feat.MelBanksOptions(num_mel_bins)
        self.mel_opts.low_freq = low_frequency
        self.mel_opts.high_freq = high_frequency
        self.mel_opts.vtln_low = vtln_low
        self.mel_opts.vtln_high = vtln_high
        self.mfcc_opts = feat.MfccOptions()
        self.mfcc_opts.frame_opts = self.frame_opts
        self.mfcc_opts.mel_opts = self.mel_opts
        self.mfcc_opts.cepstral_lifter = cepstral_lifter
        self.mfcc_opts.num_ceps = num_coefficients
        self.mfcc_opts.use_energy = use_energy
        self.mfcc_opts.energy_floor = energy_floor
        self.mfcc_opts.raw_energy = raw_energy
        self.mfcc_opts.htk_compat = htk_compatibility
        self.mfcc_obj = feat.Mfcc(self.mfcc_opts)

    def compute_mfccs(
        self,
        segment: Segment,
    ) -> np.ndarray:
        """
        Compute MFCC features for a segment

        Parameters
        ----------
        segment: :class:`~kalpy.feat.mfcc.Segment`
            Acoustic segment to generate MFCCs

        Returns
        -------
        :class:`numpy.ndarray`
            Feature matrix for the segment
        """
        duration = None
        if segment.end is not None and segment.begin is not None:
            duration = segment.end - segment.begin
        wave, sr = librosa.load(
            segment.file_path,
            sr=16000,
            offset=segment.begin,
            duration=duration,
            mono=False,
        )
        wave = np.round(wave * 32768)
        if len(wave.shape) == 2:
            channel = 0 if segment.channel is None else segment.channel
            wave = wave[channel, :]
        mfccs = self.mfcc_obj.compute(wave)
        return mfccs.numpy()

    def _compute_mfccs_for_export(
        self,
        segment: Segment,
    ) -> FloatMatrixBase:
        """
        Generate MFCCs for exporting to a kaldi archive

        Parameters
        ----------
        segment: :class:`~kalpy.feat.mfcc.Segment`
            Acoustic segment to generate MFCCs

        Returns
        -------
        :class:`_kalpy.matrix.FloatMatrixBase`
            Feature matrix for the segment
        """
        duration = None
        if segment.end is not None and segment.begin is not None:
            duration = segment.end - segment.begin
        wave, sr = librosa.load(
            segment.file_path,
            sr=16000,
            offset=segment.begin,
            duration=duration,
            mono=False,
        )
        wave = np.round(wave * 32768)
        if len(wave.shape) == 2:
            channel = 0 if segment.channel is None else segment.channel
            wave = wave[channel, :]
        mfccs = self.mfcc_obj.compute(wave)
        return mfccs

    def export_feats(
        self,
        file_name: str,
        segments: typing.Dict[str, Segment],
        write_scp: bool = False,
        compress: bool = True,
    ) -> None:
        """
        Export features to a kaldi archive file (i.e., mfccs.ark)

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Archive file path to export to
        segments: dict[str, :class:`kalpy.feat.mfcc.Segment`]
            Mapping of utterance IDs to Segment objects
        write_scp: bool
            Flag for whether an SCP file should be generated as well
        compress: bool
            Flag for whether to export features as a compressed archive
        """
        file_name = str(file_name)
        if not file_name.endswith(".ark"):
            file_name += ".ark"
        if write_scp:
            write_specifier = f"ark,scp:{file_name},{file_name.replace('.ark', '.scp')}"
        else:
            write_specifier = f"ark:{file_name}"
        if compress:
            writer = CompressedMatrixWriter(write_specifier)
        else:
            writer = BaseFloatMatrixWriter(write_specifier)
        for key, segment in segments.items():
            feats = self._compute_mfccs_for_export(segment)
            if compress:
                feats = CompressedMatrix(feats)
            writer.Write(str(key), feats)
        writer.Close()


class FeatureArchive:
    """
    Class for reading an archive or SCP of features

    Parameters
    ----------
    file_name: :class:`~pathlib.Path` or str
        Path to archive or SCP file to read from
    """

    def __init__(self, file_name: typing.Union[pathlib.Path, str]):
        self.file_name = str(file_name)
        self.read_identifier = "ark"
        if self.file_name.endswith(".scp"):
            self.read_identifier = "scp"

    def __iter__(self) -> typing.Generator[typing.Tuple[str, np.ndarray]]:
        """Iterate over the utterance features in the archive"""
        reader = SequentialBaseFloatMatrixReader(f"{self.read_identifier}:{self.file_name}")
        try:
            while not reader.Done():
                utt = reader.Key()
                feats = reader.Value().numpy()
                yield utt, feats
                reader.Next()
        finally:
            reader.Close()

    def __getitem__(self, item: str) -> np.ndarray:
        """Get features for a particular key from the archive file"""
        item = str(item)
        reader = RandomAccessBaseFloatMatrixReader(f"{self.read_identifier}:{self.file_name}")
        try:
            if not reader.HasKey(item):
                raise Exception(f"No key {item} found in {self.file_name}")
            return reader.Value(item).numpy()
        finally:
            reader.Close()



import os

import pywrapfst

from kalpy.decoder.training_graphs import TrainingGraphCompiler
from kalpy.fstext.lexicon import LexiconCompiler
from kalpy.fstext.utils import kaldi_to_pynini


def test_training_graphs(tree_path, transition_model_path, dictionary_path, temp_dir):
    lc = LexiconCompiler(position_dependent_phones=False)
    lc.load_pronunciations(dictionary_path)

    gc = TrainingGraphCompiler(transition_model_path, tree_path, lc)
    graph = kaldi_to_pynini(gc.compile_fst("this is the acoustic corpus"))
    assert graph.num_states() > 0
    assert graph.start() != pywrapfst.NO_STATE_ID

    output_file_name = temp_dir.joinpath("fsts.ark")
    gc.export_graphs(output_file_name, {"1": "this is the acoustic corpus"})
    assert output_file_name.exists()
    os.remove(output_file_name)
    gc.export_graphs(output_file_name, {"1": "this is the acoustic corpus"}, write_scp=True)
    assert output_file_name.exists()
    assert output_file_name.with_suffix(".scp").exists()



import os

from kalpy.feat.mfcc import FeatureArchive, MfccComputer, Segment


def test_generate_mfcc(wav_path):
    feature_generator = MfccComputer(snip_edges=False)
    mfccs = feature_generator.compute_mfccs(Segment(wav_path))
    assert mfccs.shape[0] == 2672
    assert mfccs.shape[1] == 13
    mfccs = feature_generator.compute_mfccs(Segment(wav_path, 1, 2, 0))
    assert mfccs.shape[0] == 100
    assert mfccs.shape[1] == 13


def test_export_mfcc(wav_path, temp_dir):
    output_file_name = temp_dir.joinpath("mfccs.ark")
    feature_generator = MfccComputer(snip_edges=False)
    segments = {"1": Segment(wav_path, 1, 2)}
    feature_generator.export_feats(output_file_name, segments)
    assert output_file_name.exists()

    archive = FeatureArchive(output_file_name)
    for utt, mfccs in archive:
        assert utt == "1"
        assert mfccs.shape[0] == 100
        assert mfccs.shape[1] == 13
    mfccs = archive["1"]
    assert mfccs.shape[0] == 100
    assert mfccs.shape[1] == 13

    os.remove(output_file_name)
    feature_generator.export_feats(output_file_name, segments, write_scp=True, compress=True)
    assert output_file_name.with_suffix(".scp").exists()

    archive = FeatureArchive(output_file_name.with_suffix(".scp"))
    for utt, mfccs in archive:
        assert utt == "1"
        assert mfccs.shape[0] == 100
        assert mfccs.shape[1] == 13
    mfccs = archive["1"]
    assert mfccs.shape[0] == 100
    assert mfccs.shape[1] == 13



"""Classes for working with lexicons"""
from __future__ import annotations

import collections
import dataclasses
import math
import pathlib
import re
import typing

import pynini
import pywrapfst


@dataclasses.dataclass
class Pronunciation:
    """
    Data class for storing information about a particular pronunciation
    """

    orthography: str
    pronunciation: str
    probability: typing.Optional[float]
    silence_after_probability: typing.Optional[float]
    silence_before_correct: typing.Optional[float]
    non_silence_before_correct: typing.Optional[float]
    disambiguation: typing.Optional[int]


def parse_dictionary_file(
    path: typing.Union[pathlib.Path, str],
) -> typing.Generator[Pronunciation]:
    """
    Parses a lexicon file and yields parsed pronunciation lines

    Parameters
    ----------
    path: :class:`~pathlib.Path` or str
        Path to lexicon file

    Yields
    ------
    str
        :class:`~kalpy.fstext.lexicon.Pronunciation`
    """
    prob_pattern = re.compile(r"\b\d+\.\d+\b")
    with open(path, encoding="utf8") as f:
        for i, line in enumerate(f):
            line = line.strip()
            if not line:
                continue
            line = line.split()
            if len(line) <= 1:
                raise Exception(
                    f'Error parsing line {i} of {path}: "{line}" did not have a pronunciation'
                )
            word = line.pop(0)
            prob = None
            silence_after_prob = None
            silence_before_correct = None
            non_silence_before_correct = None
            if prob_pattern.match(line[0]):
                prob = float(line.pop(0))
                if prob_pattern.match(line[0]):
                    silence_after_prob = float(line.pop(0))
                    if prob_pattern.match(line[0]):
                        silence_before_correct = float(line.pop(0))
                        if prob_pattern.match(line[0]):
                            non_silence_before_correct = float(line.pop(0))
            pron = " ".join(line)
            yield Pronunciation(
                word,
                pron,
                prob,
                silence_after_prob,
                silence_before_correct,
                non_silence_before_correct,
                None,
            )


class LexiconCompiler:
    """
    Class for compiling pronunciation dictionary files to lexicon FSTs

    Parameters
    ----------
    silence_disambiguation_symbol: str
        Disambiguation symbol for use in transcription decoding versus alignment decoding,
        leave empty to generate a lexicon for training/alignment
    silence_probability: float
        Probability of silence following words
    initial_silence_probability: float
        Probability of silence at the beginning of utterances
    final_silence_correction: float
        Correction factor for utterances ending in silence
    final_non_silence_correction: float
        Correction factor for utterances not ending in silence
    silence_word: str
        Word symbol to use for silence
    oov_word: str
        Word symbol to use for out of vocabulary items
    silence_phone: str
        Phone symbol to use for silence
    oov_phone: str
        Phone symbol to use for out of vocabulary items
    position_dependent_phones: bool
        Flag for using phones based on word position,
        i.e. "AA_S" (for words with a pronunciation of only "AA"),
        "AA_B" ("AA" at the beginning of a word),
        "AA_I" ("AA" internal to  a word),
        "AA_E" ("AA" at the end of a word)
        instead of using a single "AA" symbol across all word positions
    ignore_case: bool
        Flag for whether word orthographies should be transformed to lower case

    Attributes
    ----------
    word_table: :class:`pywrapfst.SymbolTable`
        Word symbol table
    phone_table: :class:`pywrapfst.SymbolTable`
        Phone symbol table
    pronunciations: list[:class:`~kalpy.fstext.lexicon.Pronunciation`]
        List of pronunciations loaded from dictionary file
    """

    def __init__(
        self,
        silence_disambiguation_symbol: typing.Optional[str] = None,
        silence_probability: float = 0.5,
        initial_silence_probability: float = 0.5,
        final_silence_correction: typing.Optional[float] = None,
        final_non_silence_correction: typing.Optional[float] = None,
        silence_word: str = "<eps>",
        oov_word: str = "<unk>",
        silence_phone: str = "sil",
        oov_phone: str = "spn",
        position_dependent_phones: bool = False,
        ignore_case: bool = True,
    ):
        self.silence_disambiguation_symbol = silence_disambiguation_symbol
        self.silence_probability = silence_probability
        self.initial_silence_probability = initial_silence_probability
        self.final_silence_correction = final_silence_correction
        self.final_non_silence_correction = final_non_silence_correction
        self.silence_word = silence_word
        self.oov_word = oov_word
        self.silence_phone = silence_phone
        self.oov_phone = oov_phone
        self.max_disambiguation_symbol = 1
        self.position_dependent_phones = position_dependent_phones
        self.ignore_case = ignore_case
        if self.silence_disambiguation_symbol is None:
            self.silence_disambiguation_symbol = "<eps>"
            self.disambiguation = False
        else:
            self.disambiguation = True
        self.word_table = pywrapfst.SymbolTable()
        self.word_table.add_symbol(silence_word)
        self.word_table.add_symbol(oov_word)
        self.phone_table = pywrapfst.SymbolTable()
        self.phone_table.add_symbol("<eps>")
        self.phone_table.add_symbol(silence_phone)
        if self.position_dependent_phones:
            for pos in ["_S", "_B", "_E", "_I"]:
                self.phone_table.add_symbol(silence_phone + pos)
        self.phone_table.add_symbol(oov_phone)
        if self.position_dependent_phones:
            for pos in ["_S", "_B", "_E", "_I"]:
                self.phone_table.add_symbol(oov_phone + pos)
        self.pronunciations: typing.List[Pronunciation] = []
        self._fst = None
        self._align_fst = None

    def to_int(self, word: str) -> int:
        """
        Look up a word in the word symbol table

        Parameters
        ----------
        word: str
            Word to look up

        Returns
        -------
        int
            Integer ID of word in symbol table
        """
        if self.word_table.member(word):
            return self.word_table.find(word)
        return self.word_table.find(self.oov_word)

    @property
    def specials_set(self) -> typing.Set[str]:
        """Special words, like the ``oov_word`` ``silence_word``, ``<s>``, and ``</s>``"""
        return {
            "#0",
            self.silence_word,
            "<s>",
            "</s>",
        }

    def load_pronunciations(self, file_name: typing.Union[pathlib.Path, str]) -> None:
        """
        Load pronunciations from a dictionary file and calculate necessary disambiguation symbols

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Path to lexicon file
        """
        non_silence_phones = set()
        words = set()
        oov_found = False
        for pron in parse_dictionary_file(file_name):
            if self.ignore_case:
                pron.orthography = pron.orthography.lower()
            if pron.orthography in self.specials_set:
                continue
            if pron.orthography == self.oov_word:
                oov_found = True
            phones = pron.pronunciation.split()
            non_silence_phones.update(phones)
            self.pronunciations.append(pron)
            words.add(pron.orthography)
        if not oov_found:
            self.pronunciations.append(
                Pronunciation(
                    self.oov_word,
                    self.oov_phone,
                    1.0,
                    None,
                    None,
                    None,
                    None,
                )
            )
        for s in sorted(words):
            self.word_table.add_symbol(s)
        for s in sorted(non_silence_phones):
            if self.position_dependent_phones:
                for pos in ["_S", "_B", "_E", "_I"]:
                    self.phone_table.add_symbol(s + pos)
            else:
                self.phone_table.add_symbol(s)

        self.compute_disambiguation_symbols()
        self.word_table.add_symbol("#0")
        self.word_table.add_symbol("<s>")
        self.word_table.add_symbol("</s>")

    @property
    def disambiguation_symbols(self) -> typing.List[int]:
        """List of integer IDs for disambiguation symbols in the phone symbol table"""
        return [
            i
            for i in range(self.phone_table.num_symbols())
            if self.phone_table.find(i).startswith("#")
        ]

    def compute_disambiguation_symbols(self):
        """Calculate the necessary disambiguation symbols for the lexicon"""
        subsequences = set()
        for pron in self.pronunciations:

            phones = pron.pronunciation.split()
            while phones:
                subsequences.add(" ".join(phones))
                phones = phones[:-1]
        last_used = collections.defaultdict(int)

        for pron in self.pronunciations:
            if pron.pronunciation in subsequences:
                last_used[pron.pronunciation] += 1
                pron.disambiguation = last_used[pron.pronunciation]

        self.max_disambiguation_symbol = max(
            self.max_disambiguation_symbol, max(last_used.values())
        )
        for x in range(self.max_disambiguation_symbol + 3):
            p = f"#{x}"
            self.phone_table.add_symbol(p)
        return self.pronunciations

    @property
    def fst(self) -> pynini.Fst:
        """Compiled lexicon FST"""
        if self._fst is not None:
            return self._fst
        initial_silence_cost = -1 * math.log(self.initial_silence_probability)
        initial_non_silence_cost = -1 * math.log(1.0 - self.initial_silence_probability)
        if self.final_silence_correction is None or self.final_non_silence_correction is None:
            final_silence_cost = 0
            final_non_silence_cost = 0
        else:
            final_silence_cost = str(-math.log(self.final_silence_correction))
            final_non_silence_cost = str(-math.log(self.final_non_silence_correction))
        base_silence_following_cost = -math.log(self.silence_probability)
        base_non_silence_following_cost = -math.log(1 - self.silence_probability)
        self.phone_table.find(self.silence_disambiguation_symbol)
        self.word_table.find("<eps>")
        self.word_table.find(self.silence_word)
        fst = pynini.Fst()
        start_state = fst.add_state()
        fst.set_start(start_state)
        non_silence_state = fst.add_state()  # Also loop state
        silence_state = fst.add_state()
        # initial no silence
        fst.add_arc(
            start_state,
            pywrapfst.Arc(
                self.phone_table.find(self.silence_disambiguation_symbol),
                self.word_table.find(self.silence_word),
                pywrapfst.Weight(fst.weight_type(), initial_non_silence_cost),
                non_silence_state,
            ),
        )
        # initial silence
        fst.add_arc(
            start_state,
            pywrapfst.Arc(
                self.phone_table.find(self.silence_phone),
                self.word_table.find(self.silence_word),
                pywrapfst.Weight(fst.weight_type(), initial_silence_cost),
                silence_state,
            ),
        )
        for pron in self.pronunciations:
            word_symbol = self.word_table.find(pron.orthography)
            phones = pron.pronunciation.split()
            silence_before_cost = (
                -math.log(pron.silence_before_correct)
                if pron.silence_before_correct is not None
                else 0.0
            )
            non_silence_before_cost = (
                -math.log(pron.non_silence_before_correct)
                if pron.non_silence_before_correct is not None
                else 0.0
            )
            silence_following_cost = (
                -math.log(pron.silence_after_probability)
                if pron.silence_after_probability is not None
                else base_silence_following_cost
            )
            non_silence_following_cost = (
                -math.log(1 - pron.silence_after_probability)
                if pron.silence_after_probability is not None
                else base_non_silence_following_cost
            )
            if self.position_dependent_phones:
                if len(phones) == 1:
                    phones[0] += "_S"
                else:
                    phones[0] += "_B"
                    phones[-1] += "_E"
                    for i in range(1, len(phones) - 1):
                        phones[i] += "_I"
            probability = pron.probability
            if probability is None:
                probability = 1
            elif probability < 0.01:
                probability = 0.01  # Dithering to ensure low probability entries
            pron_cost = abs(math.log(probability))
            if self.disambiguation and pron.disambiguation is not None:
                phones += [f"#{pron.disambiguation}"]

            new_state = fst.add_state()
            phone_symbol = self.phone_table.find(phones[0])
            # No silence before the pronunciation
            fst.add_arc(
                non_silence_state,
                pywrapfst.Arc(
                    phone_symbol,
                    word_symbol,
                    pywrapfst.Weight(fst.weight_type(), pron_cost + non_silence_before_cost),
                    new_state,
                ),
            )
            # Silence before the pronunciation
            fst.add_arc(
                silence_state,
                pywrapfst.Arc(
                    phone_symbol,
                    word_symbol,
                    pywrapfst.Weight(fst.weight_type(), pron_cost + silence_before_cost),
                    new_state,
                ),
            )
            current_state = new_state
            for i in range(1, len(phones)):
                next_state = fst.add_state()
                phone_symbol = self.phone_table.find(phones[i])
                fst.add_arc(
                    current_state,
                    pywrapfst.Arc(
                        phone_symbol,
                        self.word_table.find("<eps>"),
                        pywrapfst.Weight.one(fst.weight_type()),
                        next_state,
                    ),
                )
                current_state = next_state
            # No silence following the pronunciation
            fst.add_arc(
                current_state,
                pywrapfst.Arc(
                    self.phone_table.find(self.silence_disambiguation_symbol),
                    self.word_table.find("<eps>"),
                    pywrapfst.Weight(fst.weight_type(), non_silence_following_cost),
                    non_silence_state,
                ),
            )
            # Silence following the pronunciation
            fst.add_arc(
                current_state,
                pywrapfst.Arc(
                    self.phone_table.find(self.silence_phone),
                    self.word_table.find("<eps>"),
                    pywrapfst.Weight(fst.weight_type(), silence_following_cost),
                    silence_state,
                ),
            )
        if final_silence_cost > 0:
            fst.set_final(silence_state, pywrapfst.Weight(fst.weight_type(), final_silence_cost))
        else:
            fst.set_final(silence_state, pywrapfst.Weight.one(fst.weight_type()))
        if final_non_silence_cost > 0:
            fst.set_final(
                non_silence_state, pywrapfst.Weight(fst.weight_type(), final_non_silence_cost)
            )
        else:
            fst.set_final(non_silence_state, pywrapfst.Weight.one(fst.weight_type()))

        # fst.set_input_symbols(self.phone_table)
        # fst.set_output_symbols(self.word_table)
        fst.arcsort("olabel")
        self._fst = fst
        return self._fst

    @property
    def align_fst(self) -> pynini.Fst:
        """Compiled FST for aligning lattices when `position_dependent_phones` is False"""
        if self._align_fst is not None:
            return self._align_fst
        fst = pynini.Fst()
        start_state = fst.add_state()
        loop_state = fst.add_state()
        sil_state = fst.add_state()
        next_state = fst.add_state()
        fst.set_start(start_state)
        word_eps_symbol = self.word_table.find("<eps>")
        phone_eps_symbol = self.phone_table.find("<eps>")
        sil_cost = -math.log(0.5)
        non_sil_cost = sil_cost
        fst.add_arc(
            start_state,
            pywrapfst.Arc(
                phone_eps_symbol,
                word_eps_symbol,
                pywrapfst.Weight(fst.weight_type(), non_sil_cost),
                loop_state,
            ),
        )
        fst.add_arc(
            start_state,
            pywrapfst.Arc(
                phone_eps_symbol,
                word_eps_symbol,
                pywrapfst.Weight(fst.weight_type(), sil_cost),
                sil_state,
            ),
        )

        for pron in self.pronunciations:
            phones = pron.pronunciation.split()
            if self.position_dependent_phones:
                if phones[0] != self.silence_phone:
                    if len(phones) == 1:
                        phones[0] += "_S"
                    else:
                        phones[0] += "_B"
                        phones[-1] += "_E"
                        for i in range(1, len(phones) - 1):
                            phones[i] += "_I"
            phones = ["#1"] + phones + ["#2"]
            current_state = loop_state
            for i in range(len(phones) - 1):
                p_s = self.phone_table.find(phones[i])
                if i == 0:
                    w_s = self.word_table.find(pron.orthography)
                else:
                    w_s = word_eps_symbol
                fst.add_arc(
                    current_state,
                    pywrapfst.Arc(p_s, w_s, pywrapfst.Weight.one(fst.weight_type()), next_state),
                )
                current_state = next_state
                next_state = fst.add_state()
            i = len(phones) - 1
            if i >= 0:
                p_s = self.phone_table.find(phones[i])
            else:
                p_s = phone_eps_symbol
            if i <= 0:
                w_s = self.word_table.find(pron.orthography)
            else:
                w_s = word_eps_symbol
            fst.add_arc(
                current_state,
                pywrapfst.Arc(
                    p_s, w_s, pywrapfst.Weight(fst.weight_type(), non_sil_cost), loop_state
                ),
            )
            fst.add_arc(
                current_state,
                pywrapfst.Arc(p_s, w_s, pywrapfst.Weight(fst.weight_type(), sil_cost), sil_state),
            )
        fst.delete_states([next_state])
        fst.set_final(loop_state, pywrapfst.Weight.one(fst.weight_type()))
        fst.arcsort("olabel")
        self._align_fst = fst
        return self._align_fst



"""Classes for generating training graphs"""
import pathlib
import typing

from _kalpy.decoder import TrainingGraphCompiler as _TrainingGraphCompiler
from _kalpy.decoder import TrainingGraphCompilerOptions
from _kalpy.fstext import VectorFst, VectorFstWriter
from _kalpy.hmm import TransitionModel
from _kalpy.tree import ContextDependency
from _kalpy.util import ReadKaldiObject
from kalpy.fstext.lexicon import LexiconCompiler
from kalpy.fstext.utils import pynini_to_kaldi


class TrainingGraphCompiler:
    """
    Parameters
    ----------
    transition_model_path: str
    tree_path: str
    lexicon_compiler: :class:`~kalpy.fstext.lexicon.LexiconCompiler`
    transition_scale: float
        Scale on transitions, typically set to 0 as it will be defined during alignment
    self_loop_scale: float
        Scale on transitions, typically set to 0 as it will be defined during alignment
    batch_size: int
        Batch size for compilation, larger batches will be faster but use more memory

    Attributes
    ----------
    transition_model: :class:`~_kalpy.hmm.TransitionModel`
        Transition model to use in compiling the graphs
    tree: :class:`~_kalpy.hmm.TransitionModel`
        Context dependency tree to use in compiling the graphs
    """

    def __init__(
        self,
        transition_model_path: str,
        tree_path: str,
        lexicon_compiler: LexiconCompiler,
        transition_scale: float = 0.0,
        self_loop_scale: float = 0.0,
        batch_size: int = 250,
    ):
        self.transition_model = TransitionModel()
        ReadKaldiObject(str(transition_model_path), self.transition_model)
        self.tree = ContextDependency()
        ReadKaldiObject(str(tree_path), self.tree)
        self.lexicon_compiler = lexicon_compiler
        self.batch_size = batch_size
        kaldi_fst = pynini_to_kaldi(lexicon_compiler.fst)
        self.options = TrainingGraphCompilerOptions(transition_scale, self_loop_scale)
        self.compiler = _TrainingGraphCompiler(
            self.transition_model,
            self.tree,
            kaldi_fst,
            lexicon_compiler.disambiguation_symbols,
            self.options,
        )

    def export_graphs(
        self,
        file_name: typing.Union[pathlib.Path, str],
        transcripts: typing.Dict[str, str],
        write_scp: bool = False,
    ) -> None:
        """
        Export training graphs to a kaldi archive file (i.e., fsts.ark)

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Archive file path to export to
        transcripts: dict[str, str]
            Dictionary of utterance IDs to transcripts
        write_scp: bool
            Flag for whether an SCP file should be generated as well
        """
        file_name = str(file_name)
        if not file_name.endswith(".ark"):
            file_name += ".ark"
        if write_scp:
            write_specifier = f"ark,scp:{file_name},{file_name.replace('.ark', '.scp')}"
        else:
            write_specifier = f"ark:{file_name}"
        writer = VectorFstWriter(write_specifier)
        for key, transcript in transcripts.items():
            fst = self.compile_fst(transcript)
            writer.Write(str(key), fst)
        writer.Close()

    def compile_fst(self, transcript: str) -> VectorFst:
        """
        Compile a transcript to a training graph

        Parameters
        ----------
        transcript: str
            Orthographic transcript to compile

        Returns
        -------
        :class:`_kalpy.fstext.VectorFst`
            Training graph of transcript
        """
        transcript = [self.lexicon_compiler.to_int(x) for x in transcript.split()]
        fst = self.compiler.CompileGraphFromText(transcript)
        return fst
